// Copyright (C) 2022 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:significant

#include "qqmldebugservicefactory_p.h"

QT_BEGIN_NAMESPACE

QQmlDebugServiceFactory::~QQmlDebugServiceFactory()
    = default;

QT_END_NAMESPACE

#include "moc_qqmldebugservicefactory_p.cpp"
