// Copyright (C) 2019 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:insignificant

#ifndef QTQMLWORKERSCRIPTGLOBAL_H
#define QTQMLWORKERSCRIPTGLOBAL_H

#include <QtQml/qtqmlglobal.h>
#include <QtQmlWorkerScript/qtqmlworkerscriptexports.h>

#endif // QTQMLWORKERSCRIPTGLOBAL_H
