// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
#ifndef UNUSEDPARAM_H
#define UNUSEDPARAM_H

#include <qglobal.h>

#define UNUSED_PARAM(x) Q_UNUSED(x)

#endif // UNUSEDPARAM_H
