// Copyright 2009 the Sputnik authors.  All rights reserved.
// This code is governed by the BSD license found in the LICENSE file.

/*---
info: The Date constructor has the property "UTC"
esid: sec-date-constructor
es5id: 15.9.4_A3
description: Checking existence of the property "UTC"
---*/

if (!Date.hasOwnProperty("UTC")) {
  $ERROR('#1: The Date constructor has the property "UTC"');
}
