// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
#ifndef FASTMALLOC_H
#define FASTMALLOC_H

/* Dummy empty header file, only needed for #include source compatibility */

#endif // FASTMALLOC_H
