// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
#ifndef OWNPTR_H
#define OWNPTR_H

#include "PassOwnPtr.h"

#endif // OWNPTR_H
