// # Copyright (C) 2023 The Qt Company Ltd.
// # SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause

//! [main]
#include <QtQuickTest/quicktest.h>
#include "setup.h"

QUICK_TEST_MAIN_WITH_SETUP(TestQML, Setup)
//! [main]
